import numpy as np
from math import ceil , floor
import time
from scipy.ndimage.interpolation import shift

import matplotlib.pyplot as plt
 
#euler_maruyama integration of one-dimensional second order stochastic differential equation
#
#(1) dx = v*dt
#(2) dv = gen_force(t,x,v)*dt + \sqrt{noise_amp(t,x,v)}*dW
#
#with noise_amp(t,x,v) > 0. gen_force may include time dependent drive.
#(average output over window of size window).

def run_trajectory(dt,t_max,gen_force,noise_amp, x0 = 0, v0 = 0 , dWs = None ):
    #time-steps
    N = int(t_max/dt) 

    #initialize arrays for data
    xs = np.zeros(N + 1)
    vs = np.zeros(N + 1)

    #if no noise is passed to the solver, generate noise
    if dWs is None:
        dWs = np.random.normal(loc=0.0, scale=np.sqrt(dt), size=N+1)
    
    #initial values
    xs[0] = x0
    vs[0] = v0

    for i in range(1, N + 1):
        t = (i-1)*dt
        xs[i] = xs[i-1] + vs[i-1]*dt
        vs[i] = vs[i-1] + gen_force(t,xs[i-1],vs[i-1])*dt + np.sqrt(noise_amp(t,xs[i-1],vs[i-1]))*dWs[i-1]

    return np.array([xs,vs,dWs])

#euler_maruyama integration of general stochastic differential equation
#
# dx_i = F_i(t,x)*dt + G_ij(t,x)*dW_j
#
#with i = 0...N-1, j = 0...M-1

def run_trajectory_general(dt,t_max,F,G,x0,dWs = None):
    #time-steps
    T = int(t_max / dt)

    #learn dimensions from initial value
    N = len(F(0,x0))
    M = len(G(0,x0)[0])

    #initialize arrays for data
    xs = np.zeros([T + 1,N])

    #if no noise is passed to the solver, generate noise
    if dWs is None:
        dWs = np.random.normal(loc=0.0, scale=np.sqrt(dt), size=[T + 1,M])
    
    #initial values
    xs[0] = x0 

    for t in range(1, T + 1):
        xs[t] = xs[t-1] + F((t-1)*dt,xs[t-1])*dt + np.matmul(G((t-1)*dt,xs[t-1]), dWs[t-1])

    return [xs,dWs]



#main
def main():

    #test with brownian motion in double well potential
    t_max = 100
    dt = 0.01
    gamma = 2.2
    Temp = 0.5
    a = 1
    b = 1

    #one-dim second order solver
    gen_force = lambda t,x,v: - gamma*v - ( - a*x + b*(x**3) )
    noise_amp = lambda t,x,v: 2*Temp*gamma
    x0 = -0.1
    v0 = 0
    
    dWs = np.random.normal(loc=0.0, scale=np.sqrt(dt), size=int(t_max/dt) + 1)

    start_time = time.time()
    xs,vs,dWs = run_trajectory(dt,t_max,gen_force,noise_amp, x0 = x0, v0 = v0, dWs = dWs) 
    print("--- %s seconds ---" % (time.time() - start_time))

    #general solver for same problem: X[0] = x, X[1] = v
    F = lambda t,X: np.array([X[1] ,  - gamma*X[1] - ( - a*X[0] + b*(X[0]**3) ) ])
    G = lambda t,X: np.array([ [0] , [np.sqrt(2*Temp*gamma)] ])
    X0 = [x0,v0]

    DWs = np.array([[dW] for dW in dWs])
    
    start_time = time.time()
    Xs , DWs = run_trajectory_general(dt,t_max,F,G,X0, dWs = DWs)
    x2s, v2s = Xs.T
    print("--- %s seconds ---" % (time.time() - start_time))

    #unpack and plot
    
    
    ts = np.linspace(0, t_max , int(t_max/dt)+1)
    plt.figure(1)
    plt.plot(ts, xs  ,  color ='blue')
    plt.plot(ts, vs ,  color ='purple') 
    plt.plot(ts,  xs - x2s ,  color ='green')
    plt.plot(ts,  vs - v2s ,  color ='red') 
    #plt.figure(2)
    #plt.plot(ts,  dWs +dt ,  color ='green')
    #plt.plot(ts,  DWs[:,0] ,  color ='orange')
    plt.show(block=True)

    '''
    #ensemble average construction
    #scan over ib and calc avg. voltage
    rtj = lambda x: run_trajectory(dt,t_max,gen_force,noise_amp,x0 = x0,v0  = v0) 

    [xs,vs,dWs] = ensemble_avg(ens_size, rtj)
    '''

    return 0

if __name__== "__main__":
    main()